from .builder import ArtifactBuilder, OutdatedArtifactWarning
from .passthrough import ArtifactPassthrough
from .utilities import normalize, get_versions
from .aggregation import disaggregate, aggregate, ArtifactAggregationWarning



from typing import Set
import warnings

from gbd_mapping import causes, risk_factors, sequelae, coverage_gaps, covariates, etiologies
import pandas as pd

from vivarium_public_health.dataset_manager import EntityKey

from vivarium_inputs import core
from vivarium_inputs.data_artifact.utilities import normalize
from vivarium_inputs.utilities import normalize_for_simulation, get_age_group_bins_from_age_group_id
from vivarium_inputs.mapping_extension import healthcare_entities, health_technologies


class DataArtifactError(Exception):
    """Base error raised for issues in data artifact construction."""
    pass


class EntityError(DataArtifactError):
    """Error raised when modeled entities are improperly specified."""
    pass


CAUSE_BY_ID = {c.gbd_id: c for c in causes if c is not None}
RISK_BY_ID = {r.gbd_id: r for r in risk_factors}

AGE_COLS = ['age_group_start', 'age_group_end']
YEAR_COLS = ['year_start', 'year_end']


def loader(entity_key: EntityKey, location: str, modeled_causes: Set[str], all_measures: bool=False):
    entity_data = {
        "cause": {
            "mapping": causes,
            "getter": get_cause_data,
            "measures": ["death", "prevalence", "incidence", "cause_specific_mortality", "disability_weight",
                         "excess_mortality", "remission", "sequelae", "etiologies", "restrictions"]
        },
        "risk_factor": {
            "mapping": risk_factors,
            "getter": get_risk_data,
            "measures": ["affected_causes", "affected_risk_factors", "restrictions", "distribution",
                         "exposure_parameters", "levels", "tmred", "exposure", "exposure_standard_deviation",
                         "relative_risk", "population_attributable_fraction", "ensemble_weights"],
        },
        "sequela": {
            "mapping": sequelae,
            "getter": get_sequela_data,
            "measures": ["healthstate", "prevalence", "incidence", "disability_weight"],
        },
        "healthcare_entity": {
            "mapping": healthcare_entities,
            "getter": get_healthcare_entity_data,
            "measures": ["cost", "annual_visits"]
        },
        "health_technology": {
            "mapping": health_technologies,
            "getter": get_health_technology_data,
            "measures": ["cost", "effects", "coverage"]
        },
        "coverage_gap": {
            "mapping": coverage_gaps,
            "getter": get_coverage_gap_data,
            "measures": ["affected_causes", "affected_risk_factors", "restrictions", "distribution", "levels",
                         "relative_risk", "exposure"]
        },
        "etiology": {
            "mapping": etiologies,
            "getter": get_etiology_data,
            "measures": ["population_attributable_fraction"],
        },
        "population": {
            "mapping": {'': None},
            "getter": get_population_data,
            "measures": ["structure", "age_bins", "theoretical_minimum_risk_life_expectancy"],
        },
        "covariate": {
            "mapping": covariates,
            "getter": get_covariate_data,
            "measures": ["estimate"]
        },
        "subregions": {
            "mapping": {'': None},
            "getter": get_subregion_data,
            "measures": ["sub_region_ids"],
        },
        "dimensions": {
            "mapping": {'': None},
            "getter": get_dimension_data,
            "measures": ["full_space"]
        },
    }
    mapping, getter, measures = entity_data[entity_key.type].values()
    entity = mapping[entity_key.name]
    if not all_measures:
        return getter(entity, entity_key.measure, location, modeled_causes)
    else:
        return data_generator(entity, measures, location, modeled_causes, getter)


def data_generator(entity, measures, location, modeled_causes, getter):
    for measure in measures:
        data = getter(entity, measure, location, modeled_causes)
        if data is not None:
            yield measure, data
            del data


def get_cause_data(cause, measure, location, _):
    if measure in ["sequelae", "etiologies", "restrictions"]:
        data = get_cause_metadata(cause, measure)
    elif measure in ["death", "prevalence", "incidence", "disability_weight", "cause_specific_mortality",
                     "excess_mortality"]:
        data = core.get_draws(cause, measure, location)
        data = normalize(data)[["location", "sex", "draw", "value"] + AGE_COLS + YEAR_COLS]
    elif measure == "remission":
        data = get_cause_remission(cause, location)
    else:
        raise NotImplementedError(f"Unknown measure {measure} for cause {cause.name}")

    return data


def get_risk_data(risk, measure, location, modeled_causes):
    if measure in ["affected_causes", "affected_risk_factors", "restrictions",
                   "distribution", "exposure_parameters", "levels", "tmred"]:
        data = get_risk_metadata(risk, measure, modeled_causes)
    elif measure == "exposure":
        data = get_risk_exposure(risk, location)
    elif measure == "exposure_standard_deviation":
        data = get_risk_exposure_standard_deviation(risk, location)
    elif measure == "relative_risk":
        data = get_risk_relative_risk(risk, location)
    elif measure == "population_attributable_fraction":
        if risk.distribution not in ['normal', 'lognormal', 'ensemble']:
            raise DataArtifactError(f"PAF for {risk.name} should not be loaded from the artifact. PAF for "
                                    f"Categorical risk should be computed directly ")
        else:
            data = get_risk_population_attributable_fraction(risk, location)
    elif measure == "ensemble_weights":
        data = get_risk_ensemble_weights(risk)
    else:
        raise NotImplementedError(f"Unknown measure {measure} for risk {risk.name}")
    return data


def get_sequela_data(sequela, measure, location, _):
    if measure == "healthstate":
        data = sequela.healthstate.name
    elif measure in ["incidence", "prevalence"]:
        data = core.get_draws(sequela, measure, location).drop("sequela_id", axis=1)
        data = normalize(data)[["location", "sex", "draw", "value"] + AGE_COLS + YEAR_COLS]
        data["sequela_id"] = sequela.gbd_id
    elif measure == "disability_weight":
        data = core.get_draws(sequela, "disability_weight", location)
        data = data.drop('sex_id', 'columns')
        data = normalize(data)[['draw', 'value']]
    else:
        raise NotImplementedError(f"Unknown measure {measure} for sequela {sequela.name}")
    return data


def get_healthcare_entity_data(healthcare_entity, measure, location, _):
    if measure == "cost":
        data = core.get_draws(healthcare_entity, "cost", location)
        data = normalize(data)
        data = data.loc[data.sex == 'Male', ["location", "draw", "value"] + YEAR_COLS]
    elif measure == "annual_visits":
        data = core.get_draws(healthcare_entity, "annual_visits", location)
        data = normalize(data)
        data = data[["sex", "value", "draw"] + AGE_COLS + YEAR_COLS]
    else:
        raise NotImplementedError(f"Unknown measure {measure} for healthcare_entity {healthcare_entity.name}")
    return data


def get_health_technology_data(healthcare_technology, measure, location, _):
    if measure == "cost":
        data = core.get_draws(healthcare_technology, "cost", location)
        data = normalize(data)[["location", "draw", "value", "health_technology"] + YEAR_COLS]
    elif measure == "effects":
        data = core.get_draws(healthcare_technology, "effects", location)
        data = normalize(data)[["location", "measure", "medication", "dosage", "draw", "value"]]
    elif measure == "coverage":
        data = core.get_draws(healthcare_technology, "coverage", location)
        data = normalize(data)[AGE_COLS + YEAR_COLS
                               + ["sex", "location", "measure", "medication", "dosage", "draw", "value"]]
    else:
        raise NotImplementedError(f"Unknown measure {measure} for healthcare_entity {healthcare_technology.name}")
    return data


def get_coverage_gap_data(coverage_gap, measure, location, modeled_causes):
    if measure in ["affected_causes", "affected_risk_factors", "restrictions", "distribution", "levels"]:
        data = get_coverage_gap_metadata(coverage_gap, measure, modeled_causes)
    elif measure == "exposure":
        data = get_coverage_gap_exposure(coverage_gap, location)
    elif measure == "relative_risk":
        data = get_coverage_gap_relative_risk(coverage_gap, location)
    else:
        raise NotImplementedError(f"Unknown measure {measure} for coverage_gap {coverage_gap.name}")
    return data


def get_etiology_data(etiology, measure, location, _):
    if measure == "population_attributable_fraction":
        data = core.get_draws(etiology, "population_attributable_fraction", location)
        data = normalize(data)
        data["cause"] = data.cause_id.apply(lambda cause_id: CAUSE_BY_ID[cause_id].name)
        data = data[["location", "cause", "sex", "draw", "value", "affected_measure"] + AGE_COLS + YEAR_COLS]
    else:
        raise NotImplementedError(f"Unknown measure {measure} for etiology {etiology.name}")
    return data


def get_population_data(_, measure, location, __):
    if measure == "structure":
        data = core.get_population(location)
        data = normalize_for_simulation(data)
        data = get_age_group_bins_from_age_group_id(data)
    elif measure == "age_bins":
        data = core.get_age_bins()[["age_group_years_start", "age_group_years_end", "age_group_name"]]
        data = data.rename(columns={"age_group_years_start": "age_group_start", "age_group_years_end": "age_group_end"})
    elif measure == "theoretical_minimum_risk_life_expectancy":
        data = core.get_theoretical_minimum_risk_life_expectancy()
    else:
        raise NotImplementedError(f"Unknown measure {measure} for population.")
    return data


def get_covariate_data(covariate, measure, location, _):
    if measure == "estimate":
        data = core.get_covariate_estimate(covariate, location)
        expected_columns = ["location", "mean_value", "lower_value", "upper_value",
                            "sex_id", "year_id", "age_group_id"]

        if not set(expected_columns).issubset(data.columns):
            raise NotImplementedError(f"Unsupported covariate {covariate.name}. It does not not "
                                       f"contain the columns {expected_columns}.")
        data = data[expected_columns]
        if (data['age_group_id'] == 22).all():
            data = data.drop(['age_group_id'], axis=1)
            warnings.warn(f"Covariate \"{covariate.name}\" contains data for the age group all ages, "
                          f"so the age column is being dropped.")
        else:
            data = get_age_group_bins_from_age_group_id(data)
        data = normalize_for_simulation(data)
    else:
        raise NotImplementedError(f"Unknown or unsupported measure {measure} for covariate {covariate.name}.")
    return data


def get_subregion_data(_, measure, location, __):
    if measure == "sub_region_ids":
        data = pd.DataFrame(core.get_subregions([location]))
        data = data.melt(var_name="location", value_name="subregion_id")
    else:
        raise NotImplementedError(f"Unknown measure {measure} for subregion data.")
    return data


def get_dimension_data(_, measure, location, __):
    if measure == "full_space":
        age_bins = core.get_age_bins()
        estimation_years = core.get_estimation_years()
        data = [range(min(estimation_years), max(estimation_years) + 1),
                ["Male", "Female"], age_bins.age_group_id, [location]]
        data = pd.MultiIndex.from_product(data, names=["year", "sex", "age_group_id", "location"])
        data = data.to_frame().reset_index(drop=True)
    else:
        raise NotImplementedError(f"Unknown measure {measure} for dimensions")
    return data


##########
# Causes #
##########


def get_cause_metadata(entity, field):
    if field == "restrictions":
        data = entity.restrictions.to_dict()
    else:  # field in ["sequela", "etiologies"]:
        if entity[field] is not None:
            data = [sub_entity.name for sub_entity in entity[field]]
        else:
            data = None
    return data


def get_cause_remission(cause, location):
    result = core.get_draws(cause, "remission", location)
    result = normalize(result)[["location", "sex", "draw", "value"] + AGE_COLS + YEAR_COLS]
    return result


#########
# Risks #
#########


def get_risk_metadata(risk, measure, modeled_causes):
    if measure in ["restrictions", "exposure_parameters", "levels", "tmred"]:
        if risk[measure] is not None:
            data = risk[measure].to_dict()
        else:
            data = None
    elif measure == "affected_causes":
        data = [c.name for c in risk.affected_causes if c.name in modeled_causes]
    elif measure == "affected_risk_factors":
        # FIXME: Update mapping to include affected risks (mediation)
        data = []
    else:  # measure == "distribution"
        data = risk[measure]
    return data


def get_risk_exposure(risk, location):
    exposure = core.get_draws(risk, "exposure", location)
    result = normalize(exposure)
    result = result[["location", "sex", "draw", "value", "parameter"] + AGE_COLS + YEAR_COLS]
    return result


def get_risk_exposure_standard_deviation(risk, location):
    if risk.exposure_parameters is not None:
        exposure_sds = core.get_draws(risk, "exposure_standard_deviation", location)
        exposure_sds = normalize(exposure_sds)
        data = exposure_sds[["location", "sex", "draw", "value"] + AGE_COLS + YEAR_COLS]
    else:
        data = None
    return data


def get_risk_relative_risk(risk, location):
    rrs = core.get_draws(risk, "relative_risk", location)
    rrs = normalize(rrs)
    rrs['cause'] = rrs['cause_id'].apply(lambda cause_id: CAUSE_BY_ID[cause_id].name)
    rrs.drop(['cause_id'], axis=1, inplace=True)
    result = rrs[["location", "sex", "draw", "value", "parameter", "cause", "affected_measure"]
                 + AGE_COLS + YEAR_COLS]
    return result


def get_risk_population_attributable_fraction(risk, location):
    paf = core.get_draws(risk, 'population_attributable_fraction', location)
    paf = normalize(paf)
    paf['cause'] = paf['cause_id'].apply(lambda cause_id: CAUSE_BY_ID[cause_id].name)
    paf.drop(['cause_id'], axis=1, inplace=True)
    result = paf[["location", "sex", "draw", "value", "cause", "affected_measure"] + AGE_COLS + YEAR_COLS]
    return result


def get_risk_ensemble_weights(risk):
    if risk.distribution == "ensemble":
        weights = core.get_ensemble_weights(risk)
        weights = weights.drop(["location_id", "risk_id"], axis=1)
        weights = normalize_for_simulation(weights)
        data = get_age_group_bins_from_age_group_id(weights)
    else:
        data = None
    return data


#################
# Coverage Gaps #
#################


def get_coverage_gap_metadata(coverage_gap, measure, modeled_causes):
    if measure in ["restrictions", "levels"]:
        data = coverage_gap[measure].to_dict()
    elif measure == "affected_causes":
        data = [c.name for c in coverage_gap.affected_causes if c.name in modeled_causes]
    elif measure == "affected_risk_factors":
        data = [r.name for r in coverage_gap.affected_risk_factors]
    else:  # measure == "distribution"
        data = coverage_gap[measure]
    return data


def get_coverage_gap_exposure(coverage_gap, location):
    exposure = core.get_draws(coverage_gap, "exposure", location)
    result = normalize(exposure)
    result = result[["location", "sex", "draw", "value", "parameter"] + AGE_COLS + YEAR_COLS]
    return result


def get_coverage_gap_relative_risk(coverage_gap, location):
    data = core.get_draws(coverage_gap, "relative_risk", location)
    if data.empty:
        data = None
    else:
        data = normalize(data)
        data = data.rename(columns={'cause_id': 'cause', 'rei_id': 'risk_factor'})
        data['cause'] = data['cause'].apply(lambda c:CAUSE_BY_ID[c].name if c in CAUSE_BY_ID else c)
        data['risk_factor'] = data['risk_factor'].apply(lambda c:RISK_BY_ID[c].name if c in RISK_BY_ID else c)
        data = data[['location', 'cause', 'risk_factor', 'sex', 'draw', 'value', 'parameter', 'affected_measure'] + AGE_COLS + YEAR_COLS]
    return data



import pandas as pd
import math

from vivarium_inputs.core import get_location_id
from vivarium_inputs.utilities import normalize_for_simulation, get_age_group_bins_from_age_group_id, gbd, forecasting
from gbd_mapping import causes, covariates, etiologies
from vivarium_public_health.dataset_manager import EntityKey

NUM_DRAWS = 1000
FERTILE_AGE_GROUP_IDS = list(range(7, 15 + 1))  # need for calc live births by sex
BASE_COLUMNS = ['year_start', 'year_end', 'age_group_start', 'age_group_end', 'draw', 'sex']


def load_forecast(entity_key: EntityKey, location: str):
    entity_data = {
        "cause": {
            "mapping": causes,
            "getter": get_cause_data,
            "measures": ["cause_specific_mortality"]
        },
        "etiology": {
            "mapping": etiologies,
            "getter": get_etiology_data,
            "measures": ["incidence", "mortality"],
        },
        "population": {
            "mapping": {'': None},
            "getter": get_population_data,
            "measures": ["structure"],
        },
        "covariate": {
            "mapping": covariates,
            "getter": get_covariate_data,
            "measures": ["estimate"]
        },
    }
    mapping, getter, measures = entity_data[entity_key.type].values()
    entity = mapping[entity_key.name]
    data = getter(entity, entity_key.measure, get_location_id(location))
    data['location'] = location
    return data


def get_cause_data(cause, measure, location_id):
    data = forecasting.get_entity_measure(cause, measure, location_id)
    data = standardize_data(data, 0)
    value_column = 'value'
    data = normalize_forecasting(data, value_column)
    return data[BASE_COLUMNS + [value_column]]


def get_etiology_data(etiology, measure, location_id):
    data = forecasting.get_entity_measure(etiology, measure, location_id)
    data = standardize_data(data, 0)
    value_column = 'value'
    data = normalize_forecasting(data, value_column)
    return data[BASE_COLUMNS + [value_column]]


def get_population_data(_, measure, location_id):
    if measure == 'structure':
        data = forecasting.get_population(location_id)
        value_column = 'population'
        data = normalize_forecasting(data, value_column)
        return data[BASE_COLUMNS + [value_column]]
    else:
        raise ValueError(f"Only population.structure is supported from forecasting. You requested {measure}.")


def get_covariate_data(covariate, measure, location_id):
    if measure != 'estimate':
        raise ValueError(f"The only measure that can be retrieved for covariates is estimate. You requested {measure}.")
    value_column = 'mean_value'
    if covariate.name == 'live_births_by_sex':  # we have to calculate
        data = _get_live_births_by_sex(location_id)
    else:
        data = forecasting.get_entity_measure(covariate, measure, location_id)
        data = standardize_data(data, 0)
        data = normalize_forecasting(data, value_column)
    return data


def _get_live_births_by_sex(location_id):
    """Forecasting didn't save live_births_by_sex so have to calc from population
    and age specific fertility rate"""
    pop = forecasting.get_population(location_id)
    asfr = forecasting.get_entity_measure(covariates.age_specific_fertility_rate, 'estimate', location_id)

    # calculation of live births by sex from pop & asfr from Martin Pletcher

    fertile_pop = pop[((pop.age_group_id.isin(FERTILE_AGE_GROUP_IDS)) & (pop.sex_id == gbd.FEMALE))]
    data = asfr.merge(fertile_pop, on=['age_group_id', 'draw', 'year_id', 'sex_id', 'location_id', 'scenario'])
    data['live_births'] = data['asfr'] * data['population_agg']
    data = data.groupby(['draw', 'year_id', 'location_id'])[['live_births']].sum().reset_index()
    # normalize first because it would drop sex_id = 3 and duplicate for male and female but we need both for use in
    # vph FertilityCrudeBirthRate
    data = normalize_forecasting(data, 'mean_value')
    data['sex'] = 'Both'
    return data


def normalize_forecasting(data: pd.DataFrame, value_column='value') -> pd.DataFrame:
    assert not data.empty

    data = normalize_for_simulation(rename_value_columns(data, value_column))

    if "age_group_id" in data:
        if (data["age_group_id"] == 22).all():  # drop age if data is for all ages
            data = data.drop("age_group_id", "columns")
        else:
            # drop any age group ids that don't map to bins we use from gbd (e.g., 1 which is <5 or 158 which is <20)
            data = data[data.age_group_id.isin(gbd.get_age_bins().age_group_id)]
            data = get_age_group_bins_from_age_group_id(data)

    # not filtering on year as in vivarium_inputs.data_artifact.utilities.normalize b/c will drop future data

    if 'scenario' in data:
        data = data.drop("scenario", "columns")

    # make sure there are at least NUM_DRAWS draws
    return replicate_data(data)


def rename_value_columns(data: pd.DataFrame, value_column: str='value') -> pd.DataFrame:
    if not ('value' in data or 'mean_value' in data):
        # we need to rename the value column to match vivarium_inputs convention
        col = set(data.columns) - {'year_id', 'sex_id', 'age_group_id', 'draw', 'scenario', 'location_id'}
        if len(col) > 1:
            raise ValueError(f"You have multiple value columns in your data.")
        data = data.rename(columns={col.pop(): value_column})
    return data


def standardize_data(data: pd.DataFrame, fill_value: int) -> pd.DataFrame:
    # because forecasting data is already in long format, we need a custom standardize method

    # age_groups that we expect to exist for each entity
    whole_age_groups = gbd.get_age_group_id() if set(data.age_group_id) != {22} else [22]
    sex_id = data.sex_id.unique()
    year_id = data.year_id.unique()
    location_id = data.location_id.unique()

    index_cols = ['year_id', 'location_id', 'sex_id', 'age_group_id']
    expected_list = [year_id, location_id, sex_id, whole_age_groups]

    if 'draw' in data:
        index_cols += ['draw']
        expected_list += [data.draw.unique()]

    value_cols = {c for c in data.dropna(axis=1).columns if c not in index_cols}
    data = data.set_index(index_cols)

    # expected indexes to be in the data
    expected = pd.MultiIndex.from_product(expected_list, names=index_cols)

    new_data = data.copy()
    missing = expected.difference(data.index)

    # assign dtype=float to prevent the artifact error with mixed dtypes
    to_add = pd.DataFrame({column: fill_value for column in value_cols}, index=missing, dtype=float)

    new_data = new_data.append(to_add).sort_index()

    return new_data.reset_index()


def replicate_data(data):
    """If data has fewer than NUM_DRAWS draws, duplicate to have the full set.
    Assumes draws in data are sequential and start at 0
    """
    if 'draw' not in data:  # for things with only 1 draw, draw isn't included as a col
        data['draw'] = 0

    full_data = data.copy()
    og_draws = data.draw.unique()
    n_draws = len(og_draws)

    if n_draws < NUM_DRAWS:

        for i in range(1, math.ceil(NUM_DRAWS/n_draws)):

            max_draw = max(og_draws)
            if i == math.ceil(NUM_DRAWS/n_draws)-1 and NUM_DRAWS % n_draws > 0:
                max_draw = NUM_DRAWS % n_draws - 1

            draw_map = pd.Series(range(i*n_draws, i*n_draws + n_draws), index=og_draws)

            new_data = data[data.draw <= max_draw].copy()
            new_data['draw'] = new_data.draw.apply(lambda x: draw_map[x])

            full_data = full_data.append(new_data)

    return full_data





import logging
import warnings
import argparse
import pandas as pd
import numpy as np
from pathlib import Path
from typing import Set

from vivarium_public_health.dataset_manager import Artifact, ArtifactException, get_location_term
from vivarium_inputs.data_artifact.utilities import get_versions


_log = logging.getLogger(__name__)
METADATA = ['metadata.keyspace', 'metadata.locations', 'metadata.versions']


class ArtifactAggregationWarning(Warning):
    pass


def unpack_arguments(parser, rawargs=None):
    parser.add_argument('--config_path', type=str, required=True)
    parser.add_argument('--output_root', type=str, required=True)
    parser.add_argument('--locations', nargs='+')
    args = parser.parse_args(rawargs)

    config_path = Path(args.config_path)
    locations = args.locations
    locations = [l.replace('_', ' ') for l in locations]
    output_root = args.output_root

    return config_path, locations, output_root


def aggregate(artifacts: [Artifact], artifact_path: str) -> Artifact:
    """
    aggregates the list of artifacts with individual location. It takes a union
    of all the keys in individual artifacts. If there is any artifact which
    does not have all the keys, it will drop that location with a warning and
    aggregate only the locations with the same keys.


    Parameters
    ----------
    artifacts:
        list of Artifact objects
    artifact_path:
        path to the final aggregated artifact to be stored
    Returns
    -------
        aggregated artifact
    """

    artifact_path = Path(artifact_path).resolve()
   
    if artifact_path.is_file():
        artifact_path.unlink()

    locations = set().union(*[a.load('metadata.locations') for a in artifacts])
    keyspace_set = set().union(*[a.load('metadata.keyspace') for a in artifacts])
    valid_artifacts, valid_locations = [], []
    for a in artifacts:
        if set(a.load('metadata.keyspace')) == keyspace_set:
            valid_artifacts.append(a)
            valid_locations.extend(a.load('metadata.locations'))
        else:
            warnings.warn(f'missing_keys: {keyspace_set.difference(set(a.load("metadata.keyspace")))} '
                          f'for location:{a.load("metadata.locations")} All the artifacts not aggregated'
                          f'will be stored in {artifact_path}/broken_artifacts.', ArtifactAggregationWarning)

    artifact = Artifact(artifact_path.as_posix())
    artifact.write("metadata.locations", valid_locations)
    current_versions = get_versions()

    artifact.write("metadata.versions", current_versions)

    for k in keyspace_set - set(METADATA):
        data = [a.load(k) for a in valid_artifacts]
        if isinstance(data[0], pd.DataFrame):
            if 'location' in data[0].columns:
                data = pd.concat(data)
            else:
                assert np.all([d.equals(data[0]) for d in data])
                data = data[0]
            artifact.write(k, data)
        else:
            assert np.all([d == data[0] for d in data])
            artifact.write(k, data[0])

    # clean-up
    for loc in valid_locations:
        l = loc.replace(' ', '_').replace("'", "-")
        f = artifact_path.parent/f'{artifact_path.stem}_{l}.hdf'
        f.unlink()

    invalid_locations = set(locations).difference(set(valid_locations))
    if len(invalid_locations) > 0:
        new_dir = artifact_path.parent/'broken_artifacts'
        new_dir.mkdir()

        for loc in invalid_locations:
            l = loc.replace(' ', '_').replace("'", "-")
            f = artifact_path.parent/f'{artifact_path.stem}_{l}.hdf'
            f.rename(new_dir/f.name)

    return artifact


def main(rawargs=None):
    """ Aggregate multiple artifacts to a single artifact.
        We only take the union of the each keyspace of single artifacts and
        do not aggregate any single artifact does not have all the keys
        in the union of keyspaces.
    """
    parser = argparse.ArgumentParser()
    config_path, locations, output = unpack_arguments(parser, rawargs)

    individual_artifacts = []
    for loc in locations:
        l = loc.replace(' ', '_').replace("'", "-")
        individual_path = Path(output)/f'{config_path.stem}_{l}.hdf'
        individual_artifacts.append(Artifact(individual_path.as_posix()))

    artifact_path = f'{output}/{config_path.stem}.hdf'
    _log.debug(f'{locations}')
    if len(locations) == 1:
        location = locations[0].replace(' ', '_').replace("'", "-")
        current_artifact = Path(output) / f'{config_path.stem}_{location}.hdf'
        current_artifact.rename(Path(artifact_path))

    else:
        aggregate(individual_artifacts, artifact_path)


def disaggregate(config_name: str, output_root: str) -> Set:
    """Disaggreagte the existing multi-location artifacts into the single
    location artifacts for appending. It is only called when the append flag
    is true. If the existing artifact was built under the different versions of
    relevant libraries from the versions of those at the moment when appending
    is called, we will wipe it out and build it from scratch.

    Parameters
    ----------
    config_name:
        string of the configuration file stem i.e., existing artifact name
    output_root:
        where artifacts will be stored

    Returns
    -------
        set of existing locations
    """

    initial_artifact_path = Path(output_root) / f'{config_name}.hdf'

    if not initial_artifact_path.is_file():
        raise ArtifactException(
            f'To append it, you should provide the existing artifact. {output_root}/{config_name}.hdf does not exist')

    existing_artifact = Artifact(initial_artifact_path.as_posix())

    if existing_artifact.load('metadata.versions') != get_versions():
        #  FIXME: For now we only warn and build from scratch. We need a smarter way to handle ths.
        warnings.warn('Your artifact was built under the different versions and we cannot append it. It will be built '
                      'from scratch', ArtifactAggregationWarning)

        initial_artifact_path.unlink()
        existing_locations = {}
    else:
        existing_locations = _disaggregate(existing_artifact, initial_artifact_path)
    return existing_locations


def _disaggregate(existing_artifact: Artifact, initial_artifact_path: Path) -> Set:
    existing_keys = existing_artifact.load('metadata.keyspace')
    existing_locations = existing_artifact.load('metadata.locations')

    for loc in existing_locations:
        location = get_location_term(loc)
        l = loc.replace(' ', '_').replace("'", "-")
        temp_path = f'{initial_artifact_path.parent.as_posix()}/{initial_artifact_path.stem}_{l}.hdf'
        existing_artifact = Artifact(initial_artifact_path.as_posix(), filter_terms=[location])

        new_artifact = Artifact(temp_path)
        new_artifact.write('metadata.locations', [loc])
        new_artifact.write('metadata.versions', get_versions())

        for e_key in set(existing_keys) - set(METADATA):
            data = existing_artifact.load(e_key)
            new_artifact.write(e_key, data)

    initial_artifact_path.unlink()
    return [l.replace(' ', '_') for l in existing_locations]


if __name__ == "__main__":
    main()



from typing import Any

import pandas as pd
from vivarium_public_health.dataset_manager import EntityKey, filter_data, validate_filter_term
from vivarium_public_health.disease import DiseaseModel

from vivarium_inputs.data_artifact.loaders import loader
from vivarium_inputs.forecasting import load_forecast


class ArtifactPassthrough:

    def setup(self, builder):
        self.modeled_causes = builder.components.get_components(DiseaseModel)
        self.location = builder.configuration.input_data.location
        draw = builder.configuration.input_data.input_draw_number

        self.base_filter = {'draw': draw,
                            'location': [self.location, 'Global']}
        self.config_filter_term = validate_filter_term(builder.configuration.input_data.artifact_filter_term)

    def load(self, entity_key: str, future=False, **column_filters: str) -> Any:
        entity_key = EntityKey(entity_key)
        if future:
            data = load_forecast(entity_key, self.location)
        else:
            data = loader(entity_key, self.location, self.modeled_causes)

        if isinstance(data, pd.DataFrame):
            for key, val in self.base_filter.items():
                if key in data.columns:
                    column_filters[key] = val

            data = filter_data(data, self.config_filter_term, **column_filters)
        return data



import pytest

from pathlib import Path
from vivarium_inputs.data_artifact import ArtifactBuilder, OutdatedArtifactWarning
from vivarium_public_health.dataset_manager import EntityKey


@pytest.fixture()
def test_artifact(tmpdir):
    path = Path(tmpdir)/'test.hdf'
    location = 'United States'
    draw = 0
    artifact = ArtifactBuilder.initialize_artifact(path.as_posix(), False, draw, location)
    return artifact


def test_initialize_artifact_append_not_a_file(tmpdir):
    path = Path(tmpdir)/'test.hdf'
    location = 'United States'
    draw = 0

    with pytest.raises(ValueError):
        ArtifactBuilder.initialize_artifact(path.as_posix(), True, draw, location)


def test_initialize_artifact_append_keyspace(test_artifact):
    draw = 0
    location = 'United States'

    # artifact without a keyspace
    test_artifact.remove('metadata.keyspace')
    assert EntityKey('metadata.keyspace') not in test_artifact

    with pytest.warns(OutdatedArtifactWarning):
        artifact = ArtifactBuilder.initialize_artifact(test_artifact.path, True, draw, location)
        assert EntityKey('metadata.keyspace') in artifact


def test_initialize_artifact_append_locations(test_artifact):
    draw = 0
    location = 'United States'

    # artifact without a location
    test_artifact.remove('metadata.locations')
    assert EntityKey('metadata.locations') not in test_artifact

    with pytest.warns(OutdatedArtifactWarning):
        artifact = ArtifactBuilder.initialize_artifact(test_artifact.path, True, draw, location)
        assert EntityKey('metadata.locations') in artifact
        assert artifact.load('metadata.locations') == [location]

    # artifact with a different location
    new_location = 'Canada'
    assert artifact.load('metadata.locations') == [location]

    with pytest.raises(ValueError):
        ArtifactBuilder.initialize_artifact(test_artifact.path, True, draw, new_location)


def test_initialize_artifact_append_versions(test_artifact):
    draw = 0
    location = 'United States'

    # artifact without versions
    test_artifact.remove('metadata.versions')
    assert EntityKey('metadata.versions') not in test_artifact

    with pytest.warns(OutdatedArtifactWarning):
        artifact = ArtifactBuilder.initialize_artifact(test_artifact.path, True, draw, location)
        assert EntityKey('metadata.versions') in artifact

    current_versions = artifact.load('metadata.versions')
    new_versions = {k: '0.1' for k in current_versions}
    artifact.replace('metadata.versions', new_versions)

    assert artifact.load('metadata.versions') == new_versions

    with pytest.warns(OutdatedArtifactWarning):
        artifact = ArtifactBuilder.initialize_artifact(test_artifact.path, True, draw, location)
        assert artifact.load('metadata.versions') == current_versions


def test_initialize_artifact_no_append_file(test_artifact):
    draw = 0
    location = 'United States'

    test_artifact.write('test.key', 'data')
    assert EntityKey('test.key') in test_artifact

    # check whether the existing file wiped out
    artifact = ArtifactBuilder.initialize_artifact(test_artifact.path, False, draw, location)
    assert EntityKey('test.key') not in artifact



import hashlib
from functools import partial

import numpy as np
import pandas as pd
import pytest

from gbd_mapping.id import reiid
from gbd_mapping.cause import causes
from gbd_mapping.risk import risk_factors
from gbd_mapping.coverage_gap import coverage_gaps

from vivarium_inputs import core, utilities


# TODO there's a bunch of repeated code in the next three functions but I'm not sure what the general form should be yet
@pytest.fixture
def mock_rrs(mocker):
    rrs_mock = mocker.patch("vivarium_inputs.core._get_relative_risk")
    rr_map = {}

    def rr_builder(risks, locations):
        rrs = []
        for risk in risks:
            for cause in risk.affected_causes:
                if (risk.gbd_id, cause.gbd_id) in rr_map:
                    current_rr = rr_map[(risk.gbd_id, cause.gbd_id)]
                else:
                    current_rr = int(hashlib.md5(str((risk.gbd_id, cause.gbd_id)).encode()).hexdigest(), 16)
                    current_rr /= 2**128
                    rr_map[(risk.gbd_id, cause.gbd_id)] = current_rr+1
                age_groups = [10, 11, 12]
                years = [1990, 1995]
                sexes = [1, 2]
                idx = pd.MultiIndex.from_product(
                    [[cause.gbd_id],   [risk.gbd_id], age_groups,     years,     sexes,    locations,  ["continuous"]],
                    names=["cause_id", "risk_id",     "age_group_id", "year_id", "sex_id", "location_id", "parameter"])
                rrs.append(pd.DataFrame({f"draw_{i}": current_rr for i in range(1000)}, index=idx).reset_index())
        return pd.concat(rrs)
    rrs_mock.side_effect = rr_builder
    return rrs_mock, rr_map


@pytest.fixture
def mock_pafs(mocker, cause_list):
    risk_cause_pafs = {}

    def mock_pafs(entity_ids, location_ids):
        pafs = []
        for gbd_id in entity_ids:
            if isinstance(gbd_id, reiid):
                rids = [gbd_id]
                # TODO This assumes that all non-cause entities are diarrhea etiologies
                cids = [causes.diarrheal_diseases.gbd_id]
            else:
                rids = {r.gbd_id for r in risk_factors if gbd_id in [cc.gbd_id for cc in r.affected_causes]}
                cids = [gbd_id]
            for c in cids:
                for r in rids:
                    if risk_cause_pafs:
                        if (r, c) in risk_cause_pafs:
                            current_paf = risk_cause_pafs[(r, c)]
                        else:
                            current_paf = int(hashlib.md5(str((r, c)).encode()).hexdigest(), 16)
                            current_paf /= 2**128
                            risk_cause_pafs[(r, c)] = current_paf
                    else:
                        current_paf = 0.001
                        risk_cause_pafs[(r, c)] = current_paf

                    age_groups = [10, 11, 12]
                    years = [1990, 1995]
                    sexes = [1, 2]
                    idx = pd.MultiIndex.from_product(
                        [[c],               [r],      age_groups,     years,     sexes,    [3, 4],        location_ids],
                        names=["cause_id", "rei_id", "age_group_id", "year_id", "sex_id", "measure_id", "location_id"])
                    pafs.append(pd.DataFrame({f"draw_{i}":current_paf for i in range(1000)}, index=idx).reset_index())
        return pd.concat(pafs)

    gbd_mock = mocker.patch("vivarium_inputs.core.gbd")
    gbd_mock.get_pafs.side_effect = mock_pafs
    return gbd_mock, risk_cause_pafs


@pytest.fixture
def mock_exposures(mocker):
    exposures_mock = mocker.patch("vivarium_inputs.core._get_exposure")
    exposure_map = {}

    def exposure_builder(risks, locations):
        exposures = []
        for risk in risks:
            if risk.gbd_id in exposure_map:
                current_exposure = exposure_map[risk.gbd_id]
            else:
                current_exposure = int(hashlib.md5(str(risk.gbd_id).encode()).hexdigest(), 16)
                current_exposure /= 2**128
                current_exposure *= 100
                exposure_map[risk.gbd_id] = current_exposure
            age_groups = [10, 11, 12]
            years = [1990, 1995]
            sexes = [1, 2]
            idx = pd.MultiIndex.from_product(
                [[risk.gbd_id],   age_groups,     years,     sexes,    locations,  ["continuous"]],
                names=["risk_id", "age_group_id", "year_id", "sex_id", "location_id", "parameter"])
            exposures.append(pd.DataFrame({f"draw_{i}": current_exposure
                                           for i in range(1000)}, index=idx).reset_index())
        return pd.concat(exposures)

    exposures_mock.side_effect = exposure_builder
    return exposures_mock, exposure_map


@pytest.fixture(params=["cause", "etiology"])
def cause_like_entities(request, cause_list, etiology_list):
    if request.param == "cause":
        return cause_list
    elif request.param == "etiology":
        return etiology_list


def test_get_relative_risk(mocker):
    gbd_mock = mocker.patch("vivarium_inputs.core.gbd")
    gbd_mock_utilities = mocker.patch("vivarium_inputs.utilities.gbd")
    gbd_mock_utilities.get_age_group_id.return_value = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 30, 31, 32, 235]
    draw_cols = [f"rr_{i}" for i in range(10)]
    rr_maps = {'year_id': [1990, 1995, 2000], 'location_id': [1], 'sex_id': [1, 2], 'age_group_id': [4, 5],
               'rei_id': [240], 'cause_id': [302], 'parameter': ['cat1', 'cat2', 'cat3', 'cat4'],
               'morbidity': [1], 'mortality': [1], 'rei_id ': [240], 'modelable_entity_id': [9082], 'metric_id': [3]}

    rr_ = pd.DataFrame(columns=draw_cols, index=pd.MultiIndex.from_product([*rr_maps.values()], names=[*rr_maps.keys()]))
    rr_[draw_cols] = np.random.random_sample((len(rr_), 10)) * 10
    gbd_mock.get_relative_risk.return_value = rr_.reset_index()
    get_rr = core.get_relative_risk(risk_factors.child_wasting, 1)
    whole_age_groups = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 30, 31, 32, 235]
    missing_age_groups = list(set(whole_age_groups) - set(rr_maps['age_group_id']))
    missing_rr_maps = rr_maps.copy()
    missing_rr_maps['age_group_id'] = missing_age_groups

    del missing_rr_maps['morbidity']
    del missing_rr_maps['mortality']
    del missing_rr_maps['metric_id']
    del missing_rr_maps['modelable_entity_id']

    missing_rr = pd.DataFrame(1.0, columns=[f"draw_{i}" for i in range(10)],
                              index=pd.MultiIndex.from_product([*missing_rr_maps.values()],
                              names=[*missing_rr_maps.keys()]))
    rr_ = rr_.rename(columns={f'rr_{i}': f'draw_{i}' for i in range(10)})
    rr_ = rr_.reset_index(['morbidity', 'mortality', 'metric_id', 'modelable_entity_id'])
    rr_ = rr_.drop(['morbidity', 'mortality', 'metric_id', 'modelable_entity_id'], axis=1)
    expected_rr = rr_.append(missing_rr).sort_index().reset_index()
    expected_rr['affected_measure'] = 'incidence_rate' # every row has morbidity flag on
    get_rr = get_rr[expected_rr.columns]

    pd.util.testing.assert_frame_equal(expected_rr, get_rr)


def test_get_population_attributable_fraction(mocker):

    categorical_risk = [r for r in risk_factors if r.distribution in ['polytomous','dichotomous']]
    coverage_gap_list = [c for c in coverage_gaps]

    for risk in categorical_risk:
        with pytest.raises(core.InvalidQueryError):
            core.get_population_attributable_fraction(risk, 180)

    for cg in coverage_gap_list:
        with pytest.raises(core.InvalidQueryError):
            core.get_population_attributable_fraction(cg, 180)

    gbd_mock = mocker.patch("vivarium_inputs.core.gbd")
    gbd_mock_utilities = mocker.patch("vivarium_inputs.utilities.gbd")
    gbd_mock_utilities.get_age_group_id.return_value = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 30, 31, 32, 235]
    draw_cols = [f"draw_{i}" for i in range(10)]
    paf_maps = {'year_id': [1990, 1995, 2000], 'location_id': [1], 'sex_id': [1, 2], 'age_group_id': [4, 5],
                'rei_id': [106], 'cause_id': [493, 495], 'measure_id': [3, 4]}

    paf_ = pd.DataFrame(columns=draw_cols, index=pd.MultiIndex.from_product([*paf_maps.values()], names=[*paf_maps.keys()]))
    paf_[draw_cols] = np.random.random_sample((len(paf_), 10))
    gbd_mock.get_paf.return_value = paf_.reset_index()

    get_paf = core.get_population_attributable_fraction(risk_factors.high_total_cholesterol, 1).reset_index()

    whole_age_groups = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 30, 31, 32, 235]
    missing_age_groups = list(set(whole_age_groups) - set(paf_maps['age_group_id']))
    missing_paf_maps = paf_maps.copy()
    missing_paf_maps['age_group_id'] = missing_age_groups
    del missing_paf_maps['measure_id']

    missing_paf = pd.DataFrame(0.0, columns=draw_cols,
                               index=pd.MultiIndex.from_product([*missing_paf_maps.values(),
                                                                 ['incidence_rate', 'excess_mortality']],
                                                                names=[*missing_paf_maps.keys(), 'affected_measure'])).reset_index()
    paf_ = paf_.reset_index()
    paf_['affected_measure'] = paf_['measure_id'].apply(lambda x: 'incidence_rate' if x == 3 else 'excess_mortality')
    paf_ = paf_.drop('measure_id', 'columns')
    expected_paf = paf_.append(missing_paf).sort_index().reset_index(drop=True)
    get_paf = get_paf[expected_paf.columns]

    index_cols = ['age_group_id', 'cause_id', 'sex_id', 'year_id', 'affected_measure']
    pd.util.testing.assert_frame_equal(expected_paf.sort_values(index_cols).reset_index(drop=True),
                                       get_paf.sort_values(index_cols).reset_index(drop=True))



def mock_sequelae_disability_weights(sequelae, dw_data):
    """Returns a mock of the table of disability weights that vivarium_inputs.utilities.gbd.get_auxiliary_data returns.
     Used to patch that function call. Data is supplied to the function."""
    healthstate_id_tuples = list(map(lambda s: (s.healthstate.name, s.healthstate.gbd_id), sequelae))
    healthstates, healthstate_ids = list(zip(*healthstate_id_tuples))
    data = {'location_id': 1, 'sex_id': 3, 'age_group_id': 22, 'measure': 'disability_weight',
            'healthstate_id': healthstate_ids, 'healthstate': healthstates}

    sequelae = pd.DataFrame(data=data)
    combined = pd.concat([sequelae, dw_data], axis=1)

    return combined


def mock_sequelae_prevalence(sequela, __, prev_data):
    """Mocks sequela prevalence data, used to patch the call to get_prevalence from core.py as a side effect. Data is
    supplied to the function."""
    sequela_data = prev_data[sequela.name]
    draw_cols = [f"draw_{i}" for i in range(sequela_data.shape[1])]
    id_data = {'year_id': [1990, 1995, 2000], 'location_id': [1], 'sex_id': [1, 2],
            'age_group_id': [4, 5], 'measure_id': [5], 'metric_id': [3]}
    index = pd.DataFrame(columns=draw_cols, index=pd.MultiIndex.from_product([*id_data.values()], names=[*id_data.keys()])).index
    sequela_data.index = index
    sequela_data = sequela_data.reset_index()
    sequela_data['sequela_id'] = sequela.gbd_id

    return sequela_data


def make_test_disability_weight_data(sequelae, num_draws=10):
    """Mock data for sequelae prevalence and disability weights, as well as their weighted sum. This can be used as
    expected output for cause-level disability weights."""

    draw_cols = [f"draw_{i}" for i in range(num_draws)]
    data = {'year_id': [1990, 1995, 2000], 'location_id': [1], 'sex_id': [1, 2],
            'age_group_id': [4, 5], 'measure_id': [5], 'metric_id': [3]}
    prev_frame = pd.DataFrame(columns=draw_cols, index=pd.MultiIndex.from_product([*data.values()], names=[*data.keys()]))

    prev_data = {}  # dict is necessary based on the parameters of the fxn being mocked
    dw_data = []
    expected = None
    for s in sequelae:
        prev_draws = pd.DataFrame(data=np.random.random_sample((len(prev_frame), num_draws)), columns=draw_cols)
        dw_draws = pd.DataFrame(data=np.random.random_sample((1, num_draws)), columns=draw_cols)
        prev_data[s.name] = prev_draws
        dw_data.append(dw_draws)

        if expected is None:
            expected = pd.DataFrame(prev_draws.values * dw_draws.values)
        else:
            expected += pd.DataFrame(prev_draws.values * dw_draws.values)

    dw_data = pd.concat(dw_data, axis=0).reset_index(drop=True)
    expected.columns = [f'draw_{i}' for i in range(num_draws)]

    return prev_data, dw_data, expected


def test__get_sequela_disability_weights(mocker):
    """essentially a test of whether the aux data disability map is subset correctly."""
    cause = causes.ischemic_heart_disease

    prev_data, dw_data, expected = make_test_disability_weight_data(cause.sequelae)

    # mock disability weights
    gbd_mock = mocker.patch("vivarium_inputs.core.gbd")
    mock_disability_weight_map = mock_sequelae_disability_weights(cause.sequelae, dw_data)
    gbd_mock.get_auxiliary_data.return_value = mock_disability_weight_map

    expected = mock_disability_weight_map.loc[mock_disability_weight_map.healthstate_id == cause.sequelae[0].healthstate.gbd_id]
    actual = core._get_sequela_disability_weights(cause.sequelae[0])

    assert 'sequela_id' in actual.columns, "We assume sequela_id is included with sequela-level disability weights"
    pd.testing.assert_frame_equal(expected, actual.drop('sequela_id', 'columns'))


def test_get_disability_weight_cause(mocker):
    """Test to ensure prevalence weighted sum of disability weights is the cause-level disability weight."""

    cause = causes.urticaria

    prev_data, dw_data, expected = make_test_disability_weight_data(cause.sequelae)

    # mock disability weights
    gbd_mock = mocker.patch("vivarium_inputs.core.gbd")
    mock_disability_weight_map = mock_sequelae_disability_weights(cause.sequelae, dw_data)
    gbd_mock.get_auxiliary_data.return_value = mock_disability_weight_map

    # mock prevalence
    prevalence_mock = mocker.patch("vivarium_inputs.core.get_prevalence")
    prevalence_mock.side_effect = partial(mock_sequelae_prevalence, prev_data=prev_data)

    mock_location_id = 1.0
    cause_disability_weight = core.get_disability_weight(cause, mock_location_id)

    draw_columns = [col for col in cause_disability_weight.columns if col.startswith('draw_')]

    pd.testing.assert_frame_equal(expected, cause_disability_weight[draw_columns])




import pandas as pd
import pytest

from vivarium_inputs import forecasting


def test_rename_value_columns():
    data = pd.DataFrame({'year_id': [1990, 2000, 2010, 2020], 'scenario': 0, 'rename_me': range(0, 4)})

    assert forecasting.rename_value_columns(data).columns.tolist() == ['year_id', 'scenario', 'value']

    assert forecasting.rename_value_columns(data, 'population').columns.tolist() == ['year_id', 'scenario', 'population']

    data['new_value'] = 5

    with pytest.raises(ValueError):
        forecasting.rename_value_columns(data)


def test_normalize_forecasting():
    data = pd.DataFrame({'year_id': [1990, 2000, 2010, 2020], 'scenario': [0]*4,
                         'value': range(0, 4), 'age_group_id': 22})

    expected_base = pd.DataFrame({'year_start': [1990, 2000, 2010, 2020],
                             'year_end': [2000, 2010, 2020, 2021], 'draw': 0,
                             'value': range(0, 4)})

    expected = expected_base.copy()
    for i in range(1, forecasting.NUM_DRAWS):
        draw = expected_base.copy()
        draw['draw'] = i
        expected = expected.append(draw)

    expected = expected.set_index(['year_start', 'year_end', 'draw'])

    normalized = forecasting.normalize_forecasting(data).set_index(['year_start', 'year_end', 'draw'])

    assert normalized.equals(expected)


def test_standardize_data(mocker):
    gbd_mock_utilities = mocker.patch("vivarium_inputs.forecasting.gbd")
    gbd_mock_utilities.get_age_group_id.return_value = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
                                                        20, 30, 31, 32, 235]

    data = pd.DataFrame({'year_id': [1990, 2000, 2010, 2020], 'scenario': [0]*4,
                         'sex_id': 1, 'location_id': 102,
                         'value': range(1, 5), 'age_group_id': 2, 'draw': 1})

    standardized = forecasting.standardize_data(data, 0)

    assert 0 in standardized.value

    assert set(standardized.age_group_id) == set(forecasting.gbd.get_age_group_id())



"""Module containing functions that standardize the format of GBD outputs."""
from typing import Mapping, Iterable
from itertools import chain

import pandas as pd
import numpy as np
from gbd_mapping import (causes, risk_factors, etiologies, coverage_gaps,
                         Cause, Risk, Sequela, CoverageGap, Etiology, ModelableEntity)
from gbd_mapping.id import scalar, UNKNOWN

from vivarium_inputs.mapping_extension import HealthcareEntity, HealthTechnology

try:
    from vivarium_gbd_access import gbd
except ModuleNotFoundError:
    class GbdDummy:
        def __getattr__(self, item):
            raise ModuleNotFoundError("Required package vivarium_gbd_access not found.")
    gbd = GbdDummy()

try:
    from vivarium_gbd_access import forecasting
except ModuleNotFoundError:
    class ForecastingDummy:
        def __getattr__(self, item):
            raise ModuleNotFoundError("Required module vivarium_gbd_access.forecasting not found.")
    forecasting = ForecastingDummy()

GBD_ROUND_ID_MAP = {3: 'GBD_2015', 4: 'GBD_2016'}


class DataError(Exception):
    """Base exception for errors in data loading."""
    pass


class InvalidQueryError(DataError):
    """Exception raised when the user makes an invalid request for data (e.g. exposures for a sequela)."""
    pass


class UnhandledDataError(DataError):
    """Exception raised when we receive data from the databases that we don't know how to handle."""
    pass


class DataMissingError(DataError):
    """Exception raised when data has unhandled missing entries."""
    pass


class DuplicateDataError(DataError):
    """Exception raised when data has duplication in the index."""
    pass


def select_draw_data(data, draw, column_name, src_column=None):
    if column_name:
        if src_column is not None:
            if isinstance(src_column, str):
                column_map = {src_column.format(draw=draw): column_name}
            else:
                column_map = {src.format(draw=draw): dest for src, dest in zip(src_column, column_name)}
        else:
            column_map = {'draw_{draw}'.format(draw=draw): column_name}

        # if 'measure' is in columns, then keep it, else do
        # not keep it (need measure for the relative risk estimations)
        if 'parameter' in data.columns:
            keep_columns = ['year_id', 'age_group_start', 'age_group_end',
                            'sex_id', 'parameter'] + list(column_map.keys())
        else:
            keep_columns = ['year_id', 'age_group_start', 'age_group_end', 'sex_id'] + list(column_map.keys())

        data = data[keep_columns]
        data = data.rename(columns=column_map)

        return normalize_for_simulation(data)
    return data


def normalize_for_simulation(df):
    """
    Parameters
    ----------
    df : DataFrame
        dataframe to change

    Returns
    -------
    Returns a df with column year_id changed to year, and year_start and year_end
    created as bin ends around year_id with year_start set to year_id;
    sex_id changed to sex, and sex values changed from 1 and 2 to Male and Female

    Notes
    -----
    Used by -- load_data_from_cache

    Assumptions -- None

    Questions -- None

    Unit test in place? -- Yes
    """
    if "sex_id" in df:
        if set(df["sex_id"]) == {3}:
            df_m = df.copy()
            df_f = df.copy()
            df_m['sex'] = 'Male'
            df_f['sex'] = 'Female'
            df = pd.concat([df_m, df_f], ignore_index=True)
        else:
            df["sex"] = df.sex_id.map({1: "Male", 2: "Female", 3: "Both"}).astype(
                pd.api.types.CategoricalDtype(categories=["Male", "Female", "Both"], ordered=False))

        df = df.drop("sex_id", axis=1)

    if "year_id" in df:
        # FIXME: use central comp interpolation tools
        if 2006 in df.year_id.unique() and 2007 not in df.year_id.unique():
            df = df.loc[(df.year_id != 2006)]

        df = df.rename(columns={"year_id": "year_start"})
        idx = df.index

        mapping = df[['year_start']].drop_duplicates().sort_values(by="year_start")
        mapping['year_end'] = mapping['year_start'].shift(-1).fillna(mapping.year_start.max()+1).astype(int)

        df = df.set_index("year_start", drop=False)
        mapping = mapping.set_index("year_start", drop=False)

        df[["year_start", "year_end"]] = mapping[["year_start", "year_end"]]

        df = df.set_index(idx)

    return df


def get_age_group_bins_from_age_group_id(df):
    """Creates "age_group_start" and "age_group_end" columns from the "age_group_id" column

    Parameters
    ----------
    df: df for which you want an age column that has an age_group_id column

    Returns
    -------
    df with "age_group_start" and "age_group_end" columns
    """
    if df.empty:
        df['age_group_start'] = 0
        df['age_group_end'] = 0
        return df

    df = df.copy()
    idx = df.index
    mapping = gbd.get_age_bins()
    mapping = mapping.set_index('age_group_id')

    df = df.set_index('age_group_id')
    df[['age_group_start', 'age_group_end']] = mapping[['age_group_years_start', 'age_group_years_end']]

    df = df.set_index(idx)

    return df


def get_id_for_measure(entity: ModelableEntity, measure: str) -> int:
    """Selects the appropriate gbd id type for each entity and measure pair.

    Parameters
    ----------
    entity :
        A data containers from the `gbd_mapping` package.
    measure :
        A GBD measures requested for the provided entity.

    Returns
    -------
    The appropriate GBD id for use with central comp tools for the
    provided entity and measure.

    Raises
    ------
    InvalidQueryError
        If the entities passed are inconsistent with the requested measures.
    """
    measure_types = {
        'death': (Cause, 'gbd_id'),
        'prevalence': ((Cause, Sequela), 'gbd_id'),
        'incidence': ((Cause, Sequela), 'gbd_id'),
        'exposure': ((Risk, CoverageGap), 'gbd_id'),
        'exposure_standard_deviation': ((Risk, CoverageGap), 'gbd_id'),
        'relative_risk': ((Risk, CoverageGap), 'gbd_id'),
        'population_attributable_fraction': ((Risk, Etiology, CoverageGap), 'gbd_id'),
        'cause_specific_mortality': ((Cause,), 'gbd_id'),
        'excess_mortality': ((Cause,), 'gbd_id'),
        'annual_visits': (HealthcareEntity, 'utilization'),
        'disability_weight': (Sequela, 'gbd_id'),
        'remission': (Cause, 'dismod_id'),
        'cost': ((HealthcareEntity, HealthTechnology), 'cost'),
    }

    valid_types, id_attr = measure_types.get(measure, (None, None))

    if not valid_types or not id_attr:
        raise InvalidQueryError(f"You've requested an invalid measure: {measure}")
    elif not isinstance(entity, valid_types) or entity[id_attr] is UNKNOWN:
        raise InvalidQueryError(f"Entity {entity.name} has no data for measure '{measure}'")
    else:
        measure_id = entity[id_attr]

    return measure_id


def get_additional_id_columns(data, entity):
    id_column_map = {
        'cause': 'cause_id',
        'sequela': 'sequela_id',
        'covariate': 'covariate_id',
        'risk_factor': 'rei_id',
        'etiology': 'rei_id',
        'coverage_gap': 'coverage_gap',
        'healthcare_entity': 'healthcare_entity',
        'health_technology': 'health_technology',
    }
    out = {id_column_map[entity.kind]}
    if entity.kind == 'coverage_gap' and 'relative_risk' in data['measure'].unique():
        out.add('rei_id')
    # TODO: Why?
    out |= set(data.columns) & set(id_column_map.values())
    return out


def validate_data(data: pd.DataFrame, key_columns: Iterable[str]=None):
    """Validates that no data is missing and that the provided key columns make a valid (unique) index.

    Parameters
    ----------
    data:
        The data table to be validated.
    key_columns:
        An iterable of the column names used to uniquely identify a row in the data table.

    Raises
    ------
    DataMissingError
        If the data contains any null (NaN or NaT) values.
    DuplicatedDataError
        If the provided key columns are insufficient to uniquely identify a record in the data table.
    """

    #  check draw_cols only since we may have nan in coverage_gap data from aux_data
    draw_cols = [f'draw_{i}' for i in range(1000)]
    if np.any(data[draw_cols].isnull()):
        raise DataMissingError()

    if key_columns and np.any(data.duplicated(key_columns)):
        raise DuplicateDataError()


def get_age_group_ids(restrictions, measure='yld'):
    assert measure in ['yld', 'yll']

    age_restriction_map = {scalar(0.0): [2, None],
                           scalar(0.01): [3, 2],
                           scalar(0.10): [4, 3],
                           scalar(1.0): [5, 4],
                           scalar(5.0): [6, 5],
                           scalar(10.0): [7, 6],
                           scalar(15.0): [8, 7],
                           scalar(20.0): [9, 8],
                           scalar(30.0): [11, 10],
                           scalar(40.0): [13, 12],
                           scalar(45.0): [14, 13],
                           scalar(50.0): [15, 14],
                           scalar(55.0): [16, 15],
                           scalar(65.0): [18, 17],
                           scalar(95.0): [235, 32], }

    age_start, age_end = restrictions[f'{measure}_age_start'], restrictions[f'{measure}_age_end']
    min_age_group = age_restriction_map[age_start][0]
    max_age_group = age_restriction_map[age_end][1]
    return list(range(min_age_group, max_age_group + 1))


def get_measure_id(measure):
    name_measure_map = {'death': 1,
                        'DALY': 2,
                        'YLD': 3,
                        'YLL': 4,
                        'prevalence': 5,
                        'incidence': 6,
                        'remission': 7,
                        'excess_mortality': 9,
                        'proportion': 18,
                        'continuous': 19, }
    return name_measure_map[measure]


def standardize_all_age_groups(data: pd.DataFrame):
    if set(data.age_group_id) != {22}:
        return data
    whole_age_groups = gbd.get_age_group_id()
    missing_age_groups = set(whole_age_groups).difference(set(data.age_group_id))
    df = []
    for i in missing_age_groups:
        missing = data.copy()
        missing['age_group_id'] = i
        df.append(missing)

    return pd.concat(df, ignore_index=True)


def standardize_data(data: pd.DataFrame, fill_value: int) -> pd.DataFrame:
    # age_groups that we expect to exist for each risk
    whole_age_groups = gbd.get_age_group_id()
    sex_id = data.sex_id.unique()
    year_id = data.year_id.unique()
    location_id = data.location_id.unique()

    index_cols = ['year_id', 'location_id', 'sex_id', 'age_group_id']
    draw_cols = [c for c in data.columns if 'draw_' in c]

    other_cols = {c: data[c].unique() for c in data.dropna(axis=1).columns if c not in index_cols and c not in draw_cols}
    index_cols += [*other_cols.keys()]
    data = data.set_index(index_cols)

    # expected indexes to be in the data
    expected = pd.MultiIndex.from_product([year_id, location_id, sex_id, whole_age_groups] + [*other_cols.values()],
                                          names=(['year_id', 'location_id', 'sex_id', 'age_group_id'] + [
                                              *other_cols.keys()]))

    new_data = data.copy()
    missing = expected.difference(data.index)

    # assign dtype=float to prevent the artifact error with mixed dtypes
    to_add = pd.DataFrame({column: fill_value for column in draw_cols}, index=missing, dtype=float)

    new_data = new_data.append(to_add).sort_index()

    return new_data.reset_index()


def filter_to_most_detailed(data):
    for column, entity_list in [('cause_id', causes),
                                ('rei_id', chain(etiologies, risk_factors)),
                                ('coverage_gap_id', coverage_gaps)]:
        most_detailed = {e.gbd_id for e in entity_list if e is not None}
        if column in data:
            data = data[data[column].isin(most_detailed)]
    return data



import pandas as pd
import pkg_resources

from vivarium_inputs.utilities import normalize_for_simulation, get_age_group_bins_from_age_group_id
from vivarium_inputs import core


def normalize(data: pd.DataFrame) -> pd.DataFrame:
    """Remove GBD specific column names and concepts and make the dataframe long over draws."""
    assert not data.empty
    data = normalize_for_simulation(data)
    if "age_group_id" in data:
        data = get_age_group_bins_from_age_group_id(data)
    draw_columns = [c for c in data.columns if "draw_" in c]
    index_columns = [c for c in data.columns if "draw_" not in c]
    data = pd.melt(data, id_vars=index_columns, value_vars=draw_columns, var_name="draw")
    data["draw"] = data.draw.str.partition("_")[2].astype(int)

    estimation_years = core.get_estimation_years()
    year_start, year_end = min(estimation_years), max(estimation_years)
    if "year_start" in data:
        data = data.loc[(data.year_start >= year_start) & (data.year_end <= year_end)]
    return data


def get_versions():

    libraries = ['vivarium', 'vivarium_inputs', 'vivarium_public_health', 'gbd_mapping']
    return {k: pkg_resources.get_distribution(k).version for k in libraries}


"""This module performs the core data transformations on GBD data and provides a basic API for data access."""
from typing import Union

import pandas as pd
from gbd_mapping.base_template import ModelableEntity
from gbd_mapping.risk import Risk
from gbd_mapping.coverage_gap import CoverageGap, coverage_gaps
from gbd_mapping.cause import Cause
from gbd_mapping.sequela import Sequela

from vivarium_inputs.utilities import (InvalidQueryError, UnhandledDataError, DataMissingError,
                                       get_additional_id_columns, validate_data, get_measure_id, get_id_for_measure,
                                       standardize_all_age_groups, get_age_group_ids, filter_to_most_detailed,
                                       standardize_data, gbd)


def get_draws(entity: ModelableEntity, measure: str, location: str) -> pd.DataFrame:
    """Gets draw level gbd data for a measure and entity.

    Parameters
    ----------
    entity:
        A data container from the `gbd_mapping` package.
    measure:
        A GBD measure requested for the provided entity.
    location:
        A location to pull data for.

    Returns
    -------
    A table of draw level data for for the entity, measure, and location
    combination indexed by demographic data (age_group_id, sex_id, year_id) where appropriate.
    """
    measure_handlers = {
        'death': (get_death, set()),
        'remission': (get_remission, set()),
        'prevalence': (get_prevalence, set()),
        'incidence': (get_incidence, set()),
        'relative_risk': (get_relative_risk, {'cause_id', 'parameter', 'affected_measure'}),
        'population_attributable_fraction': (get_population_attributable_fraction, {'rei_id', 'affected_measure'}),
        'cause_specific_mortality': (get_cause_specific_mortality, set()),
        'excess_mortality': (get_excess_mortality, set()),
        'exposure': (get_exposure, {'parameter', }),
        'exposure_standard_deviation': (get_exposure_standard_deviation, {'rei_id', }),
        'annual_visits': (get_annual_visits, {'modelable_entity_id', }),
        'disability_weight': (get_disability_weight, set()),
        'cost': (get_cost, set()),
        'effects': (get_effects, {'medication', 'dosage'}),
        'coverage': (get_coverage, {'medication', 'dosage'}),
    }
    location_id = get_location_id(location)

    try:
        handler, id_columns = measure_handlers[measure]
    except KeyError:
        raise InvalidQueryError(f'No functions are available to pull data for measure {measure}')
    data = handler(entity, location_id)
    if 'measure' not in data:
        data['measure'] = measure

    id_columns |= get_additional_id_columns(data, entity)
    key_columns = ['measure'] + list(id_columns)
    for column in ['year_id', 'sex_id', 'age_group_id', 'location_id']:
        if column in data:
            key_columns.append(column)

    draw_columns = [f'draw_{i}' for i in range(0, 1000)]

    data = data[key_columns + draw_columns].reset_index(drop=True)
    validate_data(data, key_columns)
    if "location_id" in data:
        data["location"] = location
        data = data.drop("location_id", "columns")

    return data

#####################
# get_draws helpers #
#####################
#
# These functions filter out erroneous measures and deal with special cases.
#

####################
# Cause-like stuff #
####################


def get_death(entity, location_id):
    entity_id = get_id_for_measure(entity, 'death')
    death_data = gbd.get_codcorrect_draws(cause_id=entity_id, location_id=location_id)

    return death_data[death_data['measure_id'] == get_measure_id('death')]


def get_remission(entity, location_id):
    entity_id = get_id_for_measure(entity, 'remission')
    remission_data = gbd.get_modelable_entity_draws(me_id=entity_id, location_id=location_id)
    remission_data['cause_id'] = entity.gbd_id
    correct_measure = remission_data['measure_id'] == get_measure_id('remission')
    return remission_data[correct_measure]


def get_prevalence(entity, location_id):
    entity_id = get_id_for_measure(entity, 'prevalence')
    measure_data = gbd.get_como_draws(entity_id=entity_id, location_id=location_id, entity_type=entity.kind)

    correct_measure = measure_data['measure_id'] == get_measure_id('prevalence')
    correct_years = measure_data['year_id'].isin(gbd.get_estimation_years(gbd.GBD_ROUND_ID))
    return measure_data[correct_measure & correct_years]


def get_incidence(entity, location_id):
    entity_id = get_id_for_measure(entity, 'incidence')
    data = gbd.get_como_draws(entity_id=entity_id, location_id=location_id, entity_type=entity.kind)

    correct_measure = data['measure_id'] == get_measure_id('incidence')
    correct_years = data['year_id'].isin(gbd.get_estimation_years(gbd.GBD_ROUND_ID))

    # get_draws returns "total incidence." We want hazard incidence.
    # scale by the number of people not sick to convert.
    incidence = data[correct_measure & correct_years]
    prevalence = get_prevalence(entity, location_id)

    key_columns = [f'{entity.kind}_id', 'age_group_id', 'location_id', 'sex_id', 'year_id']
    draw_columns = [f'draw_{i}' for i in range(0, 1000)]

    incidence = incidence.set_index(key_columns)
    prevalence = prevalence.set_index(key_columns)

    incidence[draw_columns] = incidence[draw_columns] / (1 - prevalence[draw_columns])

    return incidence.reset_index()


def get_cause_specific_mortality(entity, location_id):
    deaths = get_draws(entity, "death", get_location_name(location_id))
    deaths["location_id"] = location_id

    population = get_population(get_location_name(location_id))
    population = population[population['year_id'] >= deaths.year_id.min()]
    population["location_id"] = location_id

    merge_columns = ['age_group_id', 'location_id', 'year_id', 'sex_id']
    key_columns = merge_columns + ['cause_id']
    draw_columns = [f'draw_{i}' for i in range(0, 1000)]

    df = deaths.merge(population, on=merge_columns).set_index(key_columns)
    csmr = df[draw_columns].divide(df['population'], axis=0).reset_index()

    csmr = csmr[key_columns + draw_columns]
    return standardize_data(csmr, 0)


def get_excess_mortality(entity, location_id):
    prevalence = get_draws(entity, 'prevalence', get_location_name(location_id)).drop('measure', 'columns')
    csmr = get_draws(entity, 'cause_specific_mortality', get_location_name(location_id)).drop('measure', 'columns')

    key_columns = ['year_id', 'sex_id', 'age_group_id', 'cause_id', 'location']
    prevalences = prevalence.set_index(key_columns)
    csmrs = csmr.set_index(key_columns)

    # In some cases CSMR is not zero for age groups where prevalence is, which leads to
    # crazy outputs. So enforce that constraint.
    # TODO: But is this the right place to do that?
    draw_columns = [f'draw_{i}' for i in range(1000)]
    csmrs[draw_columns] = csmrs[draw_columns].where(prevalences[draw_columns] != 0, 0)

    em = csmrs.divide(prevalences, axis='index').reset_index()

    em["location_id"] = location_id

    return standardize_data(em.dropna(), 0)


def _get_sequela_disability_weights(entity: Sequela):
    disability_weights = gbd.get_auxiliary_data('disability_weight', 'sequela', 'all')
    if entity.healthstate.gbd_id in disability_weights['healthstate_id'].values:
        data = disability_weights.loc[disability_weights.healthstate_id == entity.healthstate.gbd_id, :]
    else:
        id_columns = {'location_id': 1, 'age_group_id': 22, 'sex_id': 3, 'measure': 'disability_weight',
                      'healthstate_id': entity.healthstate.gbd_id, 'healthstate': entity.healthstate.name}
        draw_columns = {f'draw_{i}': 0.0 for i in range(1000)}
        id_columns.update(draw_columns)
        data = pd.DataFrame(data=id_columns, index=[0])
    data.loc[:, 'sequela_id'] = entity.gbd_id

    return data


def get_disability_weight(entity: Union[Sequela, Cause], location_id: int = None):
    if entity.kind == 'sequela':
        data = _get_sequela_disability_weights(entity)

    elif entity.kind == "cause":
        id_columns = ['age_group_id', 'year_id', 'sex_id', 'location_id']
        sequelae = entity.sequelae

        sequela_level_data = []
        for seq in sequelae:
            seq_prevalence = get_prevalence(seq, location_id).reset_index(drop=True)
            seq_disability = _get_sequela_disability_weights(seq)

            draw_columns = [col for col in seq_disability if col.startswith('draw_')]

            # duplicate disability weight to make multiplication easier.
            seq_disability = pd.concat([seq_disability[draw_columns]] * seq_prevalence.shape[0], ignore_index=True)

            seq_prevalence.loc[:, draw_columns] = seq_prevalence[draw_columns] * seq_disability
            seq_prevalence = seq_prevalence.set_index(id_columns)
            sequela_level_data.append(seq_prevalence)

        data = sum(sequela_level_data)
        data = data.drop('sequela_id', 'columns')
        data = data.reset_index()
        data.loc[:, 'cause_id'] = entity.gbd_id

    else:
        raise InvalidQueryError("Only sequela and causes have disability weights associated with them.")

    return data


###################
# Risk-like stuff #
###################

def get_relative_risk(entity: Union[Risk, CoverageGap], location_id: int):
    special_cases = [coverage_gaps.low_measles_vaccine_coverage_first_dose]
    entity_id = get_id_for_measure(entity, 'relative_risk')

    if entity.kind == 'risk_factor':
        data = pull_rr_data_from_gbd(entity_id, location_id)
        data = filter_to_most_detailed(data)
    elif entity.kind == 'coverage_gap' and entity in special_cases:
        data = pull_rr_data_from_gbd(entity_id, location_id)
        draw_cols = [f'draw_{i}' for i in range(1000)]
        data.loc[:, draw_cols] = 1 / data.loc[:, draw_cols]
        data = handle_gbd_coverage_gap_data(entity, data, 1)
        data['coverage_gap'] = entity.name
        data['affected_measure'] = 'incidence_rate' if entity.affected_causes else 'exposure_parameters'
    elif entity.kind == 'coverage_gap':
        data = gbd.get_auxiliary_data('relative_risk', 'coverage_gap', entity.name)
        data['coverage_gap'] = entity.name
        del data['measure']
        data['affected_measure'] = 'incidence_rate' if entity.affected_causes else 'exposure_parameters'
    else:
        raise InvalidQueryError(f"Entity {entity.name} has no data for measure 'relative_risk'")

    data = standardize_all_age_groups(data)
    data = standardize_data(data, 1)
    return data


def pull_rr_data_from_gbd(measure_id, location_id):
    data = gbd.get_relative_risk(risk_id=measure_id, location_id=location_id)
    data = data.rename(columns={f'rr_{i}': f'draw_{i}' for i in range(1000)})

    # FIXME: I'm passing because this is broken for zinc_deficiency, and I don't have time to investigate -J.C.
    # err_msg = ("Not all relative risk data has both the 'mortality' and 'morbidity' flag "
    #            + "set. This may not indicate an error but it is a case we don't explicitly handle. "
    #            + "If you need this risk, come talk to one of the programmers.")
    # assert np.all((measure_data.mortality == 1) & (measure_data.morbidity == 1)), err_msg
    data.loc[data.morbidity == 1, 'affected_measure'] = 'incidence_rate'
    data.loc[(data.mortality == 1) & (data.morbidity == 0), 'affected_measure'] = 'excess_mortality'

    if (((data.morbidity == 1) & (data.mortality == 0)).any()
            and not ((data.morbidity == 0) & (data.mortality == 1)).any()):
        raise DataMissingError(f"Morbidity only relative risk data for measure_id {measure_id} was retrieved, "
                               f"but no mortality only rr data found.")

    return data.drop(['metric_id', 'modelable_entity_id', 'mortality', 'morbidity'], 'columns')


def get_population_attributable_fraction(entity, location_id):
    entity_id = get_id_for_measure(entity, 'population_attributable_fraction')

    if entity.kind == 'etiology' or \
            (entity.kind == 'risk_factor' and entity.distribution in ['ensemble', 'normal', 'lognormal']):
        data = gbd.get_paf(entity_id=entity_id, location_id=location_id)
        data = filter_to_most_detailed(data)

    else:
        raise InvalidQueryError(f"Entity {entity.name} has no data for measure 'population_attributable_fraction'")

    # TODO: figure out if we need to assert some property of the different PAF measures
    mortality_data = data[data['measure_id'] == get_measure_id('YLL')]
    morbidity_data = data[data['measure_id'] == get_measure_id('YLD')]
    del mortality_data['measure_id']
    del morbidity_data['measure_id']

    # FIXME: I'm passing because this is broken for SBP, it's unimportant, and I don't have time to investigate -J.C.
    # measure_ids = {name_measure_map[m] for m in ['death', 'DALY', 'YLD', 'YLL']}
    # err_msg = ("Not all PAF data has values for deaths, DALYs, YLDs and YLLs. "
    #           + "This may not indicate an error but it is a case we don't explicitly handle. "
    #           + "If you need this PAF, come talk to one of the programmers.")
    # assert np.all(
    #    measure_data.groupby(key_columns).measure_id.unique().apply(lambda x: set(x) == measure_ids)), err_msg

    if not morbidity_data.empty and mortality_data.empty:
        raise DataMissingError(f"Morbidity population attributable fraction data for entity {entity.name} was "
                               f"retrieved, but no accompanying mortality data was found.")

    # standardize mortality and morbidity data separately because any extra cols (e.g., mort, morb) will result in dupes
    mortality_data = standardize_data(mortality_data, 0)
    morbidity_data = standardize_data(morbidity_data, 0)

    mortality_data['affected_measure'] = 'excess_mortality'
    morbidity_data['affected_measure'] = 'incidence_rate'

    data = morbidity_data
    if not mortality_data.empty: # appending if one data set is empty changes dtypes from ints to floats
        data = data.append(mortality_data)
    return data


def get_exposure(entity, location_id):
    special_cases = [coverage_gaps.low_measles_vaccine_coverage_first_dose]
    entity_id = get_id_for_measure(entity, 'exposure')

    if entity.kind in ['risk_factor', 'etiology']:
        data = gbd.get_exposure(entity_id=entity_id, location_id=location_id)
        measure_id = get_exposure_measure_id(data)
        if measure_id == get_measure_id('prevalence'):
            data = convert_exposure_prevalence_to_proportion(data)
        data = handle_exposure_from_gbd(data)
    elif entity in special_cases:
        data = gbd.get_exposure(entity_id=entity_id, location_id=location_id)
        data = handle_gbd_coverage_gap_data(entity, data, 0)
        data = handle_exposure_from_gbd(data)
        del data['modelable_entity_id']
        del data['metric_id']
    elif entity.kind == 'coverage_gap':
        data = gbd.get_auxiliary_data('exposure', entity.kind, entity.name)
        data = data[data.location_id == location_id] if set(data.location_id) != {1} else data
        data['coverage_gap'] = entity.name
        del data['measure']
    else:
        raise InvalidQueryError(f"Entity {entity.name} has no data for measure 'exposure'")

    return standardize_all_age_groups(data)


def convert_exposure_prevalence_to_proportion(data):
    # FIXME:
    # Some categorical risks come from cause models, or they get weird exposure
    # models that report prevalence instead of proportion.  We should do a
    # systematic review of them and work with the risk factors team to get the
    # exposure reported consistently.  In the mean time we scale the unit-full
    # prevalence numbers to unit-less proportion numbers. - J.C.
    key_cols = ['age_group_id', 'location_id', 'sex_id', 'year_id']
    draw_cols = [f'draw_{i}' for i in range(1000)]

    data = data.set_index(key_cols)
    data = data[draw_cols + ['rei_id', 'measure_id', 'parameter']]

    total_prevalence = data[draw_cols].reset_index().groupby(key_cols).sum()
    for parameter in data['parameter'].unique():
        correct_parameter = data['parameter'] == parameter
        data.loc[correct_parameter, draw_cols] /= total_prevalence
    data = data.reset_index()
    data['measure_id'] = get_measure_id('proportion')

    return data


def get_exposure_measure_id(data):
    measure_ids = data.measure_id.unique()
    if len(measure_ids) > 1:
        raise UnhandledDataError("Exposures should always come back with a single measure, "
                                 "or they should be dealt with as a special case.  ")

    return int(measure_ids)


def handle_exposure_from_gbd(data):
    correct_measure = data.measure_id.isin([get_measure_id('proportion'), get_measure_id('continuous')])
    data = data[correct_measure]
    # FIXME: Is this the only data we need to delete measure id for?
    del data['measure_id']
    return data


def handle_gbd_coverage_gap_data(entity, data, fill_value):
    if len(entity.affected_causes) != 1:
        raise UnhandledDataError("We only handle coverage gaps affecting a single cause. "
                                 "Tell James if you see this error.")
    restrictions = entity.affected_causes[0].restrictions
    if restrictions.yll_only:
        raise UnhandledDataError("The PAFs we use are for YLDs, so causes with no attributable YLDs should not"
                                 "have associated exposures or RRs.")

    # We pulled coverage, not exposure, so invert the categories.
    data['parameter'] = data.parameter.map({'cat1': 'cat2', 'cat2': 'cat1'})
    data['coverage_gap'] = entity.name

    age_group_ids = get_age_group_ids(restrictions, 'yld')
    correct_age_groups = data['age_group_id'].isin(age_group_ids)

    draw_cols = [f'draw_{i}' for i in range(1000)]
    data.loc[~correct_age_groups, draw_cols] = fill_value

    return data


def get_exposure_standard_deviation(entity, location_id):
    entity_id = get_id_for_measure(entity, 'exposure_standard_deviation')
    data = gbd.get_exposure_standard_deviation(entity_id, location_id)

    key_cols = ['age_group_id', 'location_id', 'sex_id', 'year_id', 'rei_id']
    draw_cols = [f'draw_{i}' for i in range(1000)]

    return data[key_cols + draw_cols]


###############
# Other stuff #
###############

def get_annual_visits(entity, location_id):
    entity_id = get_id_for_measure(entity, 'annual_visits')
    measure_data = gbd.get_modelable_entity_draws(me_id=entity_id, location_id=location_id)
    measure_data['healthcare_entity'] = entity.name

    correct_measure = measure_data['measure_id'] == get_measure_id('continuous')
    correct_sex = measure_data['sex_id'] != gbd.COMBINED

    return measure_data[correct_measure & correct_sex]


def get_cost(entity, location_id):
    cost_data = gbd.get_auxiliary_data('cost', entity.kind, entity.name)
    if 'location_id' in cost_data.columns:
        cost_data = cost_data[cost_data['location_id'] == location_id]
    return cost_data


def get_effects(entity, _):
    data = gbd.get_auxiliary_data('effects', entity.kind, entity.name)
    data = data.drop(columns=['year_id', 'sex_id'])
    return data


def get_coverage(entity, location_id):
    data = gbd.get_auxiliary_data('coverage', entity.kind, entity.name)
    data = data[data['location_id'] == location_id]
    return data


####################################
# Measures for risk like entities  #
####################################


def get_ensemble_weights(risk):
    data = gbd.get_auxiliary_data('ensemble_weights', 'risk_factor', 'all')
    data = data[data['risk_id'] == risk.gbd_id]
    return data


#######################
# Other kinds of data #
#######################


def get_population(location):
    population = gbd.get_population(get_location_id(location))
    population["location"] = location
    keep_columns = ['age_group_id', 'location', 'year_id', 'sex_id', 'population']
    return population[keep_columns]


def get_age_bins():
    return gbd.get_age_bins()


def get_theoretical_minimum_risk_life_expectancy():
    data = gbd.get_theoretical_minimum_risk_life_expectancy()
    # TODO: Figure out a more rigorous way to handle generating age bins
    # (esp. last edge) from precise ages
    data = data.rename(columns={'age': 'age_group_start'})
    data['age_group_end'] = data.age_group_start.shift(-1).fillna(125)
    return data


def get_subregions(location):
    return gbd.get_subregions(get_location_id(location))


def get_covariate_estimate(covariate, location):
    location_id = get_location_id(location)
    data = gbd.get_covariate_estimate(covariate.gbd_id, location_id)
    data['location'] = location
    data = data.drop('location_id', 'columns')
    return data


def get_location_id(location_name):
    return {r.location_name: r.location_id for _, r in gbd.get_location_ids().iterrows()}[location_name]


def get_location_name(location_id):
    return {r.location_id: r.location_name for _, r in gbd.get_location_ids().iterrows()}[location_id]


def get_estimation_years():
    return gbd.get_estimation_years(gbd.GBD_ROUND_ID)




